""" some utility functions """

import os
import string


def is_hex_string(hex_string):
    """ check if the passed in string is really hex """
    if hex_string is None:
        return False
    return all(c in string.hexdigits for c in hex_string)


def is_valid_file(filepath):
    """ check if the passed filepath points to a real file """
    if filepath is None:
        return False
    return os.path.isfile(filepath)


def get_x_bits(num, max_bits, num_bits, right_bits=True):
    """ ensure the correct number of bits and pull the upper x bits """
    bits = bin(num).lstrip("0b")
    bits = bits.zfill(max_bits)
    if right_bits:
        return int(bits[-num_bits:], 2)
    return int(bits[:num_bits], 2)



""" Cuckoo Filter, python implementation
    License: MIT
    Author: Tyler Barrus (barrust@gmail.com)
"""

import os
import random
from numbers import Number
from struct import calcsize, pack, unpack

from ..exceptions import CuckooFilterFullError, InitializationError
from ..hashes import fnv_1a
from ..utilities import get_x_bits, is_valid_file


class CuckooFilter(object):
    """ Simple Cuckoo Filter implementation

        Args:
            capacity (int): The number of bins
            bucket_size (int): The number of buckets per bin
            max_swaps (int): The number of cuckoo swaps before stopping
            expansion_rate (int): The rate at which to expand
            auto_expand (bool): If the filter should automatically expand
            finger_size (int): The size of the fingerprint to use in bytes \
            (between 1 and 4); exported as 4 bytes; up to the user to reset \
            the size correctly on import
            filename (str): The path to the file to load or None if no file
            hash_function (function): Hashing strategy function to use \
            `hf(key)`
        Returns:
            CuckooFilter: A Cuckoo Filter object """

    __slots__ = [
        "_bucket_size",
        "_cuckoo_capacity",
        "__max_cuckoo_swaps",
        "__expansion_rate",
        "__auto_expand",
        "__fingerprint_size",
        "__hash_func",
        "_inserted_elements",
        "_buckets",
    ]

    def __init__(
        self,
        capacity=10000,
        bucket_size=4,
        max_swaps=500,
        expansion_rate=2,
        auto_expand=True,
        finger_size=4,
        filepath=None,
        hash_function=None,
    ):
        """ setup the data structure """
        valid_prms = (
            isinstance(capacity, Number)
            and capacity >= 1
            and isinstance(bucket_size, Number)
            and bucket_size >= 1
            and isinstance(max_swaps, Number)
            and max_swaps >= 1
        )
        if not valid_prms:
            msg = "CuckooFilter: capacity, bucket_size, and max_swaps " "must be an integer greater than 0"
            raise InitializationError(msg)
        self._bucket_size = int(bucket_size)
        self._cuckoo_capacity = int(capacity)
        self.__max_cuckoo_swaps = int(max_swaps)
        self.__expansion_rate = None
        self.expansion_rate = expansion_rate
        self.__auto_expand = None
        self.auto_expand = auto_expand
        self.__fingerprint_size = None
        self.fingerprint_size = finger_size

        if hash_function is None:
            self.__hash_func = fnv_1a
        else:
            self.__hash_func = hash_function
        self._inserted_elements = 0
        if filepath is None:
            self._buckets = list()
            for _ in range(self.capacity):
                self.buckets.append(list())
        elif is_valid_file(filepath):
            self._load(filepath)
        else:
            msg = "CuckooFilter: failed to load provided file"
            raise InitializationError(msg)

    def __contains__(self, key):
        """ setup the `in` keyword """
        return self.check(key)

    def __str__(self):
        """ setup what it will print """
        msg = (
            "{0}:\n"
            "\tCapacity: {1}\n"
            "\tTotal Bins: {2}\n"
            "\tLoad Factor: {3}%\n"
            "\tInserted Elements: {4}\n"
            "\tMax Swaps: {5}\n"
            "\tExpansion Rate: {6}\n"
            "\tAuto Expand: {7}"
        )
        return msg.format(
            self.__class__.__name__,
            self.capacity,
            self.capacity * self.bucket_size,
            self.load_factor() * 100,
            self.elements_added,
            self.max_swaps,
            self.expansion_rate,
            self.auto_expand,
        )

    @property
    def elements_added(self):
        """int: The number of elements added

        Note:
            Not settable"""
        return self._inserted_elements

    @property
    def capacity(self):
        """int: The number of bins

        Note:
            Not settable"""
        return self._cuckoo_capacity

    @property
    def max_swaps(self):
        """int: The maximum number of swaps to perform

        Note:
            Not settable"""
        return self.__max_cuckoo_swaps

    @property
    def bucket_size(self):
        """int: The number of buckets per bin

        Note:
            Not settable"""
        return self._bucket_size

    @property
    def buckets(self):
        """list(list): The buckets holding the fingerprints

        Note:
            Not settable"""
        return self._buckets

    @property
    def expansion_rate(self):
        """ int: The rate at expansion when the filter grows """
        return self.__expansion_rate

    @expansion_rate.setter
    def expansion_rate(self, val):
        """ set the self expand value """
        self.__expansion_rate = int(val)

    @property
    def auto_expand(self):
        """ bool: True if the cuckoo filter will expand automatically """
        return self.__auto_expand

    @auto_expand.setter
    def auto_expand(self, val):
        """ set the self expand value """
        self.__auto_expand = bool(val)

    @property
    def fingerprint_size(self):
        """int: The size in bytes of the fingerprint

        Raises:
            ValueError: If the size is not between 1 and 4
        Note:
            The size of the fingerprint must be between 1 and 4"""
        return self.__fingerprint_size // 8

    @fingerprint_size.setter
    def fingerprint_size(self, val):
        """ set the fingerprint size """
        tmp = int(val)
        if not 1 <= tmp <= 4:
            msg = ("{}: fingerprint size must be between 1 " "and 4").format(self.__class__.__name__)
            raise ValueError(msg)
        self.__fingerprint_size = tmp * 8  # bytes to bits

    def load_factor(self):
        """ float: How full the Cuckoo Filter is currently """
        return self.elements_added / (self.capacity * self.bucket_size)

    def add(self, key):
        """ Add element key to the filter

            Args:
                key (str): The element to add
            Raises:
                CuckooFilterFullError: When element not inserted after \
                maximum number of swaps or 'kicks' """
        idx_1, idx_2, fingerprint = self._generate_fingerprint_info(key)

        is_present = self._check_if_present(idx_1, idx_2, fingerprint)
        if is_present is not None:  # already there, nothing to do
            return
        finger = self._insert_fingerprint(fingerprint, idx_1, idx_2)
        self._deal_with_insertion(finger)

    def check(self, key):
        """Check if an element is in the filter

        Args:
            key (str): Element to check
        Returns:
            bool: True if likely present, False if definately not"""
        idx_1, idx_2, fingerprint = self._generate_fingerprint_info(key)
        is_present = self._check_if_present(idx_1, idx_2, fingerprint)
        if is_present is not None:
            return True
        return False

    def remove(self, key):
        """Remove an element from the filter

        Args:
            key (str): Element to remove
        Returns:
            bool: True if removed, False if not present"""
        idx_1, idx_2, fingerprint = self._generate_fingerprint_info(key)
        idx = self._check_if_present(idx_1, idx_2, fingerprint)
        if idx is None:
            return False
        self.buckets[idx].remove(fingerprint)
        self._inserted_elements -= 1
        return True

    def export(self, filename):
        """Export cuckoo filter to file

        Args:
            filename (str): Path to file to export"""
        with open(filename, "wb") as filepointer:
            for bucket in self.buckets:
                # do something for each...
                rep = len(bucket) * "I"
                filepointer.write(pack(rep, *bucket))
                leftover = self.bucket_size - len(bucket)
                rep = leftover * "I"
                filepointer.write(pack(rep, *([0] * leftover)))
            # now put out the required information at the end
            filepointer.write(pack("II", self.bucket_size, self.max_swaps))

    def expand(self):
        """ Expand the cuckoo filter """
        self._expand_logic(None)

    def _insert_fingerprint(self, fingerprint, idx_1, idx_2):
        """ insert a fingerprint """
        if self.__insert_element(fingerprint, idx_1):
            self._inserted_elements += 1
            return None
        elif self.__insert_element(fingerprint, idx_2):
            self._inserted_elements += 1
            return None

        # we didn't insert, so now we need to randomly select one index to use
        # and move things around to the other index, if possible, until we
        # either move everything around or hit the maximum number of swaps
        idx = random.choice([idx_1, idx_2])

        for _ in range(self.max_swaps):
            # select one element to be swapped out...
            swap_elm = random.randint(0, self.bucket_size - 1)

            swb = self.buckets[idx][swap_elm]
            fingerprint, self.buckets[idx][swap_elm] = swb, fingerprint

            # now find another place to put this fingerprint
            index_1, index_2 = self._indicies_from_fingerprint(fingerprint)

            idx = index_2 if idx == index_1 else index_1

            if self.__insert_element(fingerprint, idx):
                self._inserted_elements += 1
                return None

        # if we got here we have an error... we might need to know what is left
        return fingerprint

    def _load(self, filename):
        """ load a cuckoo filter from file """
        with open(filename, "rb") as filepointer:
            offset = calcsize("II")
            int_size = calcsize("I")
            filepointer.seek(offset * -1, os.SEEK_END)
            list_size = filepointer.tell()
            mybytes = unpack("II", filepointer.read(offset))
            self._bucket_size = mybytes[0]
            self.__max_cuckoo_swaps = mybytes[1]
            self._cuckoo_capacity = list_size // int_size // self.bucket_size
            self._inserted_elements = 0
            # now pull everything in!
            filepointer.seek(0, os.SEEK_SET)
            self._buckets = list()
            for i in range(self.capacity):
                self.buckets.append(list())
                for _ in range(self.bucket_size):
                    fingerprint = unpack("I", filepointer.read(int_size))[0]
                    if fingerprint != 0:
                        self.buckets[i].append(fingerprint)
                        self._inserted_elements += 1

    def _check_if_present(self, idx_1, idx_2, fingerprint):
        """ wrapper for checking if fingerprint is already inserted """
        if fingerprint in self.buckets[idx_1]:
            return idx_1
        elif fingerprint in self.buckets[idx_2]:
            return idx_2
        return None

    def __insert_element(self, fingerprint, idx):
        """ insert element wrapper """
        if len(self.buckets[idx]) < self.bucket_size:
            self.buckets[idx].append(fingerprint)
            return True
        return False

    def _expand_logic(self, extra_fingerprint):
        """ the logic to acutally expand the cuckoo filter """
        # get all the fingerprints
        fingerprints = self._setup_expand(extra_fingerprint)

        for finger in fingerprints:
            idx_1, idx_2 = self._indicies_from_fingerprint(finger)
            res = self._insert_fingerprint(finger, idx_1, idx_2)
            if res is not None:  # again, this *shouldn't* happen
                msg = "The CuckooFilter failed to expand"
                raise CuckooFilterFullError(msg)

    def _setup_expand(self, extra_fingerprint):
        """ setup this thing """
        fingerprints = list()
        if extra_fingerprint is not None:
            fingerprints.append(extra_fingerprint)
        for idx in range(self.capacity):
            fingerprints.extend(self.buckets[idx])

        self._cuckoo_capacity = self.capacity * self.expansion_rate
        self._buckets = list()
        self._inserted_elements = 0
        for _ in range(self.capacity):
            self.buckets.append(list())

        return fingerprints

    def _indicies_from_fingerprint(self, fingerprint):
        """ Generate the possible insertion indicies from a fingerprint

            Args:
                fingerprint (int): The fingerprint to use for generating \
                indicies """
        idx_1 = fingerprint % self.capacity
        idx_2 = self.__hash_func(str(fingerprint)) % self.capacity
        return idx_1, idx_2

    def _generate_fingerprint_info(self, key):
        """Generate the fingerprint and indicies using the provided key

        Args:
            key (str): The element for which information is to be generated
        """
        # generate the fingerprint along with the two possible indecies
        hash_val = self.__hash_func(key)
        fingerprint = get_x_bits(hash_val, 64, self.__fingerprint_size, True)
        idx_1, idx_2 = self._indicies_from_fingerprint(fingerprint)

        # NOTE: This should never happen...
        if idx_1 > self.capacity or idx_2 > self.capacity:
            msg = "Either idx_1 {0} or idx_2 {1} is greater than {2}"
            raise ValueError(msg.format(idx_1, idx_2, self.capacity))
        return idx_1, idx_2, fingerprint

    def _deal_with_insertion(self, finger):
        """ some code to handle the insertion the same """
        if finger is None:
            return
        elif self.auto_expand:
            self._expand_logic(finger)
        else:
            msg = "The {} is currently full".format(self.__class__.__name__)
            raise CuckooFilterFullError(msg)



""" BloomFilter, python implementation
    License: MIT
    Author: Tyler Barrus (barrust@gmail.com)
"""

import math
import os
from abc import abstractmethod
from binascii import hexlify, unhexlify
from itertools import chain
from numbers import Number
from struct import Struct, calcsize, pack, unpack
from textwrap import wrap

from ..exceptions import InitializationError
from ..hashes import default_fnv_1a
from ..utilities import is_hex_string, is_valid_file


class BaseBloom(object):
    """ basic bloom filter object """

    __slots__ = [
        "_bloom",
        "__num_bits",
        "__est_elements",
        "__fpr",
        "__number_hashes",
        "__hash_func",
        "_els_added",
        "_on_disk",
        "__impt_type",
        "__blm_type",
        "__bloom_length",
    ]

    def __init__(
        self,
        blm_type,
        est_elements=None,
        false_positive_rate=None,
        filepath=None,
        hex_string=None,
        hash_function=None,
    ):
        """ setup the basic values needed """
        self._bloom = None
        self.__num_bits = 0  # number of bits
        self.__est_elements = est_elements
        self.__fpr = 0.0
        self.__number_hashes = 0
        self.__hash_func = default_fnv_1a
        self._els_added = 0
        self._on_disk = False  # not on disk
        self.__blm_type = blm_type
        if self.__blm_type in ["regular", "reg-ondisk", "expanding"]:
            self.__impt_type = "B"
        else:
            self.__impt_type = "I"

        if blm_type in ["regular", "reg-ondisk", "expanding"]:
            msg = "Insufecient parameters to set up the Bloom Filter"
        else:
            msg = "Insufecient parameters to set up the Counting Bloom " "Filter"

        if is_valid_file(filepath):
            self.__load(blm_type, filepath, hash_function)
        elif is_hex_string(hex_string):
            self._load_hex(hex_string, hash_function)
        elif est_elements is not None and false_positive_rate is not None:
            vals = self._set_optimized_params(est_elements, float(false_positive_rate), hash_function)
            self.__hash_func = vals[0]
            self.__fpr = vals[1]
            self.__number_hashes = vals[2]
            self.__num_bits = vals[3]
            if blm_type in ["regular", "reg-ondisk"]:
                self.__bloom_length = int(math.ceil(self.__num_bits / 8.0))
            else:
                self.__bloom_length = self.number_bits
            if blm_type not in ["reg-ondisk"]:
                self._bloom = [0] * self.bloom_length
        else:
            raise InitializationError(msg)

    def __contains__(self, key):
        """ setup the `in` keyword """
        return self.check(key)

    def clear(self):
        """ Clear or reset the Counting Bloom Filter """
        self._els_added = 0
        for idx in range(self.bloom_length):
            self._bloom[idx] = 0

    @property
    def false_positive_rate(self):
        """float: The maximum desired false positive rate

        Note:
            Not settable"""
        return self.__fpr

    @property
    def estimated_elements(self):
        """int: The maximum number of elements estimated to be added at setup

        Note:
            Not settable"""
        return self.__est_elements

    @property
    def number_hashes(self):
        """int: The number of hashes required for the Bloom Filter hashing
        strategy

        Note:
            Not settable"""
        return self.__number_hashes

    @property
    def number_bits(self):
        """int: Number of bits in the Bloom Filter

        Note:
            Not settable"""
        return self.__num_bits

    @property
    def elements_added(self):
        """ int: Number of elements added to the Bloom Filter

        Note:
            Changing this can cause the current false positive rate to \
            be reported incorrectly """
        return self._els_added

    @elements_added.setter
    def elements_added(self, val):
        """ set the els added """
        self._els_added = val

    @property
    def is_on_disk(self):
        """bool: Is the Bloom Filter on Disk or not

        Note:
            Not settable"""
        return self._on_disk

    @property
    def bloom_length(self):
        """int: Length of the Bloom Filter array

        Note:
            Not settable"""
        return self.__bloom_length

    @property
    def bloom(self):
        """ list(int): The bit/int array """
        return self._bloom

    @property
    def hash_function(self):
        """function: The hash function used

        Note:
            Not settable"""
        return self.__hash_func

    def hashes(self, key, depth=None):
        """ Return the hashes based on the provided key

            Args:
                key (str): Description of arg1
                depth (int): Number of permutations of the hash to generate; \
                if None, generate `number_hashes`
            Returns:
                List(int): A list of the hashes for the key in int form """
        tmp = depth if depth is not None else self.number_hashes
        return self.__hash_func(key, tmp)

    @staticmethod
    def _set_optimized_params(estimated_elements, false_positive_rate, hash_function):
        """ set the parameters to the optimal sizes """
        if hash_function is None:
            tmp_hash = default_fnv_1a
        else:
            tmp_hash = hash_function

        valid_prms = isinstance(estimated_elements, Number) and estimated_elements > 0
        if not valid_prms:
            msg = "Bloom: estimated elements must be greater than 0"
            raise InitializationError(msg)
        valid_prms = isinstance(false_positive_rate, Number) and 0.0 <= false_positive_rate < 1.0
        if not valid_prms:
            msg = "Bloom: false positive rate must be between 0.0 and 1.0"
            raise InitializationError(msg)

        fpr = pack("f", float(false_positive_rate))
        t_fpr = unpack("f", fpr)[0]  # to mimic the c version!
        # optimal caluclations
        n_els = estimated_elements
        m_bt = math.ceil((-n_els * math.log(t_fpr)) / 0.4804530139182)  # ln(2)^2
        number_hashes = int(round(0.6931471805599453 * m_bt / n_els))  # math.log(2.0)

        if number_hashes <= 0:  # this should never happen...
            msg = "Bloom: Number hashes is zero; unusable parameters provided"
            raise InitializationError(msg)

        return tmp_hash, t_fpr, number_hashes, int(m_bt)

    def __load(self, blm_type, filename, hash_function=None):
        """ load the Bloom Filter from file """
        # read in the needed information, and then call _set_optimized_params
        # to set everything correctly
        with open(filename, "rb") as filepointer:
            offset = calcsize("QQf")
            filepointer.seek(offset * -1, os.SEEK_END)
            mybytes = unpack("QQf", filepointer.read(offset))
            vals = self._set_optimized_params(mybytes[0], mybytes[2], hash_function)
            self.__hash_func = vals[0]
            self.__fpr = vals[1]
            self.__number_hashes = vals[2]
            self.__num_bits = vals[3]
            self.__est_elements = mybytes[0]
            self._els_added = mybytes[1]
            if blm_type in ["regular", "reg-ondisk"]:
                self.__bloom_length = int(math.ceil(self.__num_bits / 8.0))
            else:
                self.__bloom_length = self.number_bits
            # now read in the bit array!
            filepointer.seek(0, os.SEEK_SET)
            offset = calcsize(self.__impt_type) * self.bloom_length
            rep = self.__impt_type * self.bloom_length
            self._bloom = list(unpack(rep, filepointer.read(offset)))

    def _load_hex(self, hex_string, hash_function=None):
        """ placeholder for loading from hex string """
        offset = calcsize(">QQf") * 2
        stct = Struct(">QQf")
        tmp_data = stct.unpack_from(unhexlify(hex_string[-offset:]))
        vals = self._set_optimized_params(tmp_data[0], tmp_data[2], hash_function)
        self.__hash_func = vals[0]
        self.__fpr = vals[1]
        self.__number_hashes = vals[2]
        self.__num_bits = vals[3]
        self.__est_elements = tmp_data[0]
        self._els_added = tmp_data[1]
        if self.__blm_type in ["regular", "reg-ondisk"]:
            self.__bloom_length = int(math.ceil(self.__num_bits / 8.0))
        else:
            self.__bloom_length = self.number_bits

        tmp_bloom = unhexlify(hex_string[:-offset])
        rep = self.__impt_type * self.bloom_length
        self._bloom = list(unpack(rep, tmp_bloom))

    def export_hex(self):
        """Export the Bloom Filter as a hex string

        Return:
            str: Hex representation of the Bloom Filter"""
        mybytes = pack(
            ">QQf",
            self.estimated_elements,
            self.elements_added,
            self.false_positive_rate,
        )
        if self.__blm_type in ["regular", "reg-ondisk"]:
            bytes_string = hexlify(bytearray(self.bloom)) + hexlify(mybytes)
        else:
            bytes_string = b""
            for val in self.bloom:
                bytes_string += hexlify(pack(self.__impt_type, val))
            bytes_string += hexlify(mybytes)
        return str(bytes_string, "utf-8")

    def export(self, filename):
        """ Export the Bloom Filter to disk

            Args:
                filename (str): The filename to which the Bloom Filter will \
                be written. """
        with open(filename, "wb") as filepointer:
            rep = self.__impt_type * self.bloom_length
            filepointer.write(pack(rep, *self.bloom))
            filepointer.write(
                pack(
                    "QQf",
                    self.estimated_elements,
                    self.elements_added,
                    self.false_positive_rate,
                )
            )

    def export_c_header(self, filename):
        """ Export the Bloom Filter to disk as a C header file.

            Args:
                filename (str): The filename to which the Bloom Filter will \
                be written. """
        trailer = pack(
            ">QQf",
            self.estimated_elements,
            self.elements_added,
            self.false_positive_rate,
        )
        data = ("  " + line for line in wrap(", ".join(("0x{:02x}".format(e) for e in chain(self.bloom, trailer))), 80))
        with open(filename, "w") as file:
            print("#include <inttypes.h>", file=file)
            print("const uint64_t estimated_elements = ", self.estimated_elements, ";", sep="", file=file)
            print("const uint64_t elements_added = ", self.elements_added, ";", sep="", file=file)
            print("const float false_positive_rate = ", self.false_positive_rate, ";", sep="", file=file)
            print("const uint64_t number_bits = ", self.number_bits, ";", sep="", file=file)
            print("const unsigned int number_hashes = ", self.number_hashes, ";", sep="", file=file)
            print("const unsigned char bloom[] = {", *data, "};", sep="\n", file=file)

    def export_size(self):
        """Calculate the size of the bloom on disk

        Returns:
            int: Size of the Bloom Filter when exported to disk"""
        tmp_b = calcsize(self.__impt_type)
        return (self.bloom_length * tmp_b) + calcsize("QQf")

    def current_false_positive_rate(self):
        """Calculate the current false positive rate based on elements added

        Return:
            float: The current false positive rate"""
        num = self.number_hashes * -1 * self.elements_added
        dbl = num / float(self.number_bits)
        exp = math.exp(dbl)
        return math.pow((1 - exp), self.number_hashes)

    def estimate_elements(self):
        """Estimate the number of unique elements added

        Returns:
            int: Number of elements estimated to be inserted"""
        setbits = self._cnt_number_bits_set()
        log_n = math.log(1 - (float(setbits) / float(self.number_bits)))
        tmp = float(self.number_bits) / float(self.number_hashes)
        return int(-1 * tmp * log_n)

    @staticmethod
    def __cnt_set_bits(i):
        """ count number of bits set in this int """
        return bin(i).count("1")

    def _cnt_number_bits_set(self):
        """ calculate the total number of set bits in the bloom """
        setbits = 0
        for i in list(range(0, self.bloom_length)):
            setbits += self.__cnt_set_bits(self._get_element(i))
        return setbits

    def _get_element(self, idx):
        """ wrappper for getting an element from the Bloom Filter! """
        return self._bloom[idx]

    def add(self, key):
        """Add the key to the Bloom Filter

        Args:
            key (str): The element to be inserted"""
        hashes = self.hashes(key)
        self.add_alt(hashes)

    def add_alt(self, hashes):
        """ Add the element represented by hashes into the Bloom Filter

            Args:
                hashes (list): A list of integers representing the key to \
                insert """
        for i in list(range(0, self.number_hashes)):
            k = int(hashes[i]) % self.number_bits
            idx = k // 8
            j = self._get_element(idx)
            tmp_bit = int(j) | int((1 << (k % 8)))
            self._bloom[idx] = tmp_bit
        self._els_added += 1

    def check(self, key):
        """Check if the key is likely in the Bloom Filter

        Args:
            key (str): The element to be checked
        Returns:
            bool: True if likely encountered, False if definately not"""
        hashes = self.hashes(key)
        return self.check_alt(hashes)

    def check_alt(self, hashes):
        """ Check if the element represented by hashes is in the Bloom Filter

            Args:
                hashes (list): A list of integers representing the key to \
                check
            Returns:
                bool: True if likely encountered, False if definately not """
        for i in list(range(0, self.number_hashes)):
            k = int(hashes[i]) % self.number_bits
            if (int(self._get_element(k // 8)) & int((1 << (k % 8)))) == 0:
                return False
        return True

    @abstractmethod
    def union(self, second):
        """ Return a new Bloom Filter that contains the union of the two """
        pass

    @abstractmethod
    def intersection(self, second):
        """Return a new Bloom Filter that contains the intersection of the
        two"""
        pass

    @abstractmethod
    def jaccard_index(self, second):
        """ Return a the Jaccard Similarity score between two bloom filters """
        pass

    def _verify_bloom_similarity(self, second):
        """ can the blooms be used in intersection, union, or jaccard index """
        hash_match = self.number_hashes != second.number_hashes
        same_bits = self.number_bits != second.number_bits
        next_hash = self.hashes("test") != second.hashes("test")
        if hash_match or same_bits or next_hash:
            return False
        return True



""" BloomFilter, python implementation
    License: MIT
    Author: Tyler Barrus (barrust@gmail.com)
    URL: https://github.com/barrust/pyprobables
"""

import os
from struct import calcsize, pack, unpack

from ..exceptions import RotatingBloomFilterError
from ..utilities import is_valid_file
from .bloom import BloomFilter


class ExpandingBloomFilter(object):
    """ Simple expanding Bloom Filter implementation for use in python; the
        Bloom Fiter will automatically expand, or grow, if the false
        positive rate is about to become greater than the desired false
        positive rate.

        Args:
            est_elements (int): The number of estimated elements to be added
            false_positive_rate (float): The desired false positive rate
            filepath (str): Path to file to load
            hash_function (function): Hashing strategy function to use \
            `hf(key, number)`
        Returns:
            ExpandingBloomFilter: An expanding Bloom Filter object
        Note:
            Initialization order of operations:
                1) Filepath
                2) est_elements and false_positive_rate"""

    __slots__ = [
        "_blooms",
        "__fpr",
        "__est_elements",
        "__hash_func",
        "__added_elements",
    ]

    def __init__(
        self,
        est_elements=None,
        false_positive_rate=None,
        filepath=None,
        hash_function=None,
    ):
        """ initialize """
        self._blooms = list()
        self.__fpr = false_positive_rate
        self.__est_elements = est_elements
        self.__hash_func = hash_function
        self.__added_elements = 0  # total added...

        if is_valid_file(filepath):
            self.__load(filepath)
        else:
            # add in the initial bloom filter!
            self.__add_bloom_filter()

    def __contains__(self, key):
        """ setup the `in` functionality """
        return self.check(key)

    @property
    def expansions(self):
        """ int: The number of expansions """
        return len(self._blooms) - 1

    @property
    def false_positive_rate(self):
        """ float: The desired false positive rate of the expanding Bloom \
                   Filter """
        return self.__fpr

    @property
    def estimated_elements(self):
        """int: The original number of elements estimated to be in the Bloom \
                Filter """
        return self.__est_elements

    @property
    def elements_added(self):
        """ int: The total number of elements added """
        return self.__added_elements

    def push(self):
        """ Push a new expansion onto the Bloom Filter """
        self.__add_bloom_filter()

    def check(self, key):
        """ Check to see if the key is in the Bloom Filter

            Args:
                key (str): The key to check for in the Bloom Filter
            Returns:
                bool: `True` if the element is likely present; `False` if \
                      definately not present """
        hashes = self._blooms[0].hashes(key)
        return self.check_alt(hashes)

    def check_alt(self, hashes):
        """ Check to see if the hashes are in the Bloom Filter

            Args:
                hashes (list): The hash representation to check for in the \
                               Bloom Filter
            Returns:
                bool: `True` if the element is likely present; `False` if \
                      definately not present """
        for blm in self._blooms:
            if blm.check_alt(hashes):
                return True
        return False

    def add(self, key, force=False):
        """ Add the key to the Bloom Filter

            Args:
                key (str): The element to be inserted
                force (bool): `True` will force it to be inserted, even if it \
                              likely has been inserted before `False` will \
                              only insert if not found in the Bloom Filter """
        hashes = self._blooms[0].hashes(key)
        self.add_alt(hashes, force)

    def add_alt(self, hashes, force=False):
        """ Add the element represented by hashes into the Bloom Filter

            Args:
                hashes (list): A list of integers representing the key to insert
                force (bool): `True` will force it to be inserted, even if \
                              it likely has been inserted before \
                              `False` will only insert if not found in the \
                              Bloom Filter """
        self.__added_elements += 1
        if force or not self.check_alt(hashes):
            self.__check_for_growth()
            self._blooms[-1].add_alt(hashes)

    def __add_bloom_filter(self):
        """ build a new bloom and add it on! """
        blm = BloomFilter(
            est_elements=self.__est_elements,
            false_positive_rate=self.__fpr,
            hash_function=self.__hash_func,
        )
        self._blooms.append(blm)

    def __check_for_growth(self):
        """ detereming if the bloom filter should automatically grow """
        if self._blooms[-1].elements_added >= self.__est_elements:
            self.__add_bloom_filter()

    def export(self, filepath):
        """Export an expanding Bloom Filter, or subclass, to disk

        Args:
            filepath (str): The path to the file to import"""
        with open(filepath, "wb") as fileobj:
            # add all the different Bloom bit arrays...
            for blm in self._blooms:
                rep = "Q" + "B" * blm.bloom_length
                fileobj.write(pack(rep, blm.elements_added, *blm.bloom))
            fileobj.write(
                pack(
                    "QQQf",
                    len(self._blooms),
                    self.estimated_elements,
                    self.elements_added,
                    self.false_positive_rate,
                )
            )

    def __load(self, filename):
        """ load a file """
        with open(filename, "rb") as fileobj:
            offset = calcsize("QQQf")
            fileobj.seek(offset * -1, os.SEEK_END)
            size, est_els, els_added, fpr = unpack("QQQf", fileobj.read(offset))

            fileobj.seek(0, os.SEEK_SET)
            # set the basic defaults
            self._blooms = list()
            self.__added_elements = els_added
            self.__fpr = fpr
            self.__est_elements = est_els
            for _ in range(size):
                blm = BloomFilter(
                    est_elements=self.__est_elements,
                    false_positive_rate=self.__fpr,
                    hash_function=self.__hash_func,
                )
                # now we need to read in the correct number of bytes...
                offset = calcsize("Q") + calcsize("B") * blm.bloom_length
                rep = "Q" + "B" * blm.bloom_length
                unpacked = list(unpack(rep, fileobj.read(offset)))
                blm._bloom = unpacked[1:]
                blm.elements_added = unpacked[0]
                self._blooms.append(blm)


class RotatingBloomFilter(ExpandingBloomFilter):
    """ Simple Rotating Bloom Filter implementation that allows for the "older"
        elements added to be removed, in chunks. As the queue fills up, those
        elements inserted earlier will be bulk removed. This also provides the
        user with the oportunity to force the removal instead of it being time
        based.

        Args:
            est_elements (int): The number of estimated elements to be added
            false_positive_rate (float): The desired false positive rate
            max_queue_size (int): This is the number is used to determine the \
            maximum number of Bloom Filters. Total elements added is based on \
            `max_queue_size * est_elements`
            filepath (str): Path to file to load
            hash_function (function): Hashing strategy function to use \
            `hf(key, number)`
        Note:
            Initialization order of operations:
                1) Filepath
                2) est_elements and false_positive_rate
    """

    __slots__ = [
        "_blooms",
        "__fpr",
        "__est_elements",
        "__hash_func",
        "__added_elements",
        "_queue_size",
    ]

    def __init__(
        self,
        est_elements=None,
        false_positive_rate=None,
        max_queue_size=10,
        filepath=None,
        hash_function=None,
    ):
        """ initialize """
        super(RotatingBloomFilter, self).__init__(
            est_elements=est_elements,
            false_positive_rate=false_positive_rate,
            filepath=filepath,
            hash_function=hash_function,
        )
        self._queue_size = max_queue_size
        self.__added_elements = self.elements_added
        self.__est_elements = self.estimated_elements
        self.__fpr = self.false_positive_rate
        self.__hash_func = hash_function

    @property
    def max_queue_size(self):
        """ int: The maximum size for the queue """
        return self._queue_size

    @property
    def current_queue_size(self):
        """ int: The current size of the queue """
        return len(self._blooms)

    def add_alt(self, hashes, force=False):
        """ Add the element represented by hashes into the Bloom Filter

            Args:
                hashes (list): A list of integers representing the key to insert
                force (bool): `True` will force it to be inserted, even if \
                              it likely has been inserted before \
                              `False` will only insert if not found in the \
                              Bloom Filter """
        self.__added_elements += 1
        if force or not self.check_alt(hashes):
            self.__rotate_bloom_filter()
            self._blooms[-1].add_alt(hashes)

    def pop(self):
        """Pop the oldest Bloom Filter off of the queue without pushing a new
        Bloom Filter onto the queue

        Raises:
            RotatingBloomFilterError: Unable to rotate the Bloom Filter"""
        if self.current_queue_size == 1:
            msg = "Popping a Bloom Filter will result in an unusable system!"
            raise RotatingBloomFilterError(msg)
        self._blooms.pop(0)

    def push(self):
        """ Push a new bloom filter onto the queue and rotate if necessary """
        self.__rotate_bloom_filter(force=True)

    def __rotate_bloom_filter(self, force=False):
        """handle determining if/when the Bloom Filter queue needs to be
        rotated"""
        blm = self._blooms[-1]
        ready_to_rotate = blm.elements_added == blm.estimated_elements
        no_need_to_pop = self.current_queue_size < self._queue_size
        if force and no_need_to_pop:
            self.__add_bloom_filter()
        elif force:  # must need to be pop'd first!
            blm = self._blooms.pop(0)
            self.__add_bloom_filter()
        elif ready_to_rotate and no_need_to_pop:
            self.__add_bloom_filter()
        elif ready_to_rotate:
            blm = self._blooms.pop(0)
            self.__add_bloom_filter()

    def __add_bloom_filter(self):
        """ build a new bloom and add it on! """
        blm = BloomFilter(
            est_elements=self.__est_elements,
            false_positive_rate=self.__fpr,
            hash_function=self.__hash_func,
        )
        self._blooms.append(blm)



""" Count-Min Sketch python implementation
    License: MIT
    Author: Tyler Barrus (barrust@gmail.com)
    URL: https://github.com/barrust/count-min-sketch
"""

import math
import os
from numbers import Number
from struct import calcsize, pack, unpack

from ..constants import INT32_T_MAX, INT32_T_MIN, INT64_T_MAX, INT64_T_MIN
from ..exceptions import CountMinSketchError, InitializationError, NotSupportedError
from ..hashes import default_fnv_1a
from ..utilities import is_valid_file


class CountMinSketch(object):
    """ Simple Count-Min Sketch implementation for use in python;
        It can read and write the same format as the c version
        (https://github.com/barrust/count-min-sketch)

        Args:
            width (int): The width of the count-min sketch
            depth (int): The depth of the count-min sketch
            confidence (float): The level of confidence desired
            error_rate (float): The desired error rate
            filepath (str): Path to file to load
            hash_function (function): Hashing strategy function to use \
            `hf(key, number)`
        Returns:
            CountMinSketch: A Count-Min Sketch object

        Note:
            Initialization order of operations:
                1) From file
                2) Width and depth
                3) Confidence and error rate
        Note:
            Default query type is `min`
        Note:
            For width and depth, width may realistically be in the thousands \
            while depth is in the single digit to teens """

    __slots__ = [
        "__width",
        "__depth",
        "__confidence",
        "__error_rate",
        "__elements_added",
        "__query_method",
        "_bins",
        "_hash_function",
    ]

    def __init__(
        self,
        width=None,
        depth=None,
        confidence=None,
        error_rate=None,
        filepath=None,
        hash_function=None,
    ):
        """ default initilization function """
        # default values
        self.__width = 0
        self.__depth = 0
        self.__confidence = 0.0
        self.__error_rate = 0.0
        self.__elements_added = 0
        self.__query_method = self.__min_query

        if is_valid_file(filepath):
            self.__load(filepath)
        elif width is not None and depth is not None:
            valid_prms = isinstance(width, Number) and width > 0 and isinstance(depth, Number) and depth > 0
            if not valid_prms:
                msg = "CountMinSketch: width and depth must be greater than 0"
                raise InitializationError(msg)
            self.__width = int(width)
            self.__depth = int(depth)
            self.__confidence = 1 - (1 / math.pow(2, self.depth))
            self.__error_rate = 2 / self.width
            self._bins = [0] * (self.width * self.depth)
        elif confidence is not None and error_rate is not None:
            valid_prms = (
                isinstance(confidence, Number) and confidence > 0 and isinstance(error_rate, Number) and error_rate > 0
            )
            if not valid_prms:
                msg = "CountMinSketch: width and depth must be greater than 0"
                raise InitializationError(msg)
            self.__confidence = confidence
            self.__error_rate = error_rate
            self.__width = math.ceil(2 / error_rate)
            numerator = -1 * math.log(1 - confidence)
            self.__depth = math.ceil(numerator / 0.6931471805599453)
            self._bins = [0] * int(self.width * self.depth)
        else:
            msg = (
                "Must provide one of the following to initialize the "
                "Count-Min Sketch:\n"
                "    A file to load,\n"
                "    The width and depth,\n"
                "    OR confidence and error rate"
            )
            raise InitializationError(msg)

        if hash_function is None:
            self._hash_function = default_fnv_1a
        else:
            self._hash_function = hash_function

    def __str__(self):
        """ string representation of the count min sketch """
        msg = (
            "Count-Min Sketch:\n"
            "\tWidth: {0}\n"
            "\tDepth: {1}\n"
            "\tConfidence: {2}\n"
            "\tError Rate: {3}\n"
            "\tElements Added: {4}"
        )
        return msg.format(
            self.width,
            self.depth,
            self.confidence,
            self.error_rate,
            self.elements_added,
        )

    def __contains__(self, key):
        """ setup the `in` keyword """
        return True if self.check(key) != 0 else False

    @property
    def width(self):
        """int: The width of the count-min sketch

        Note:
            Not settable"""
        return self.__width

    @property
    def depth(self):
        """int: The depth of the count-min sketch

        Note:
            Not settable"""
        return self.__depth

    @property
    def confidence(self):
        """float: The confidence of the count-min sketch

        Note:
            Not settable"""
        return self.__confidence

    @property
    def error_rate(self):
        """float: The error rate of the count-min sketch

        Note:
            Not settable"""
        return self.__error_rate

    @property
    def elements_added(self):
        """int: The number of elements added to the count-min sketch

        Note:
            Not settable"""
        return self.__elements_added

    @property
    def query_type(self):
        """str: The name of the query type being used

        Note:
            Valid values:
                * 'min' or None
                * 'mean'
                * 'mean-min'"""
        if self.__query_method == self.__mean_query:
            return "mean"
        elif self.__query_method == self.__mean_min_query:
            return "mean-min"
        return "min"

    @query_type.setter
    def query_type(self, val):
        """set to min query Options='min', 'mean', 'mean-min'
        other values are set to min
        setting to mean is converting to a Count-Mean Sketch
        setting to mean-min is converting to a Count-Mean-Min Sketch"""
        if val is None:
            self.__query_method = self.__min_query
            return
        val = val.lower()
        if val == "mean":
            self.__query_method = self.__mean_query
        elif val == "mean-min":
            self.__query_method = self.__mean_min_query
        else:
            self.__query_method = self.__min_query

    def clear(self):
        """ Reset the count-min sketch to an empty state """
        self.__elements_added = 0
        for i, _ in enumerate(self._bins):
            self._bins[i] = 0

    def hashes(self, key, depth=None):
        """ Return the hashes based on the provided key

            Args:
                key (str): Description of arg1
                depth (int): Number of permutations of the hash to generate; \
                if None, generate `number_hashes`

            Returns:
                List(int): A list of the hashes for the key in int form """
        t_depth = self.depth if depth is None else depth
        return self._hash_function(key, t_depth)

    def add(self, key, num_els=1):
        """ Insert the element `key` into the count-min sketch

            Args:
                key (str): The element to insert
                num_els (int): The number of times to insert the element
            Returns:
                int: The number of times the element was likely inserted \
                after the insertion """
        hashes = self.hashes(key)
        return self.add_alt(hashes, num_els)

    def add_alt(self, hashes, num_els=1):
        """ Insert an element by using the hash representation

            Args:
                key (str): The element to insert
                num_els (int): The number of times to insert the element
            Returns:
                int: The number of times the element was likely inserted \
                after the insertion """
        res = list()
        for i, val in enumerate(hashes):
            t_bin = (val % self.width) + (i * self.width)
            self._bins[t_bin] += num_els
            if self._bins[t_bin] > INT32_T_MAX:
                self._bins[t_bin] = INT32_T_MAX
            res.append(self._bins[t_bin])
        self.__elements_added += num_els

        if self.elements_added > INT64_T_MAX:
            self.__elements_added = INT64_T_MAX
        return self.__query_method(sorted(res))

    def remove(self, key, num_els=1):
        """ Remove element 'key' from the count-min sketch

            Args:
                key (str): The element to remove
                num_els (int): The number of times to remove the element
            Returns:
                int: The number of times the element was likely inserted \
                after the removal """
        hashes = self.hashes(key)
        return self.remove_alt(hashes, num_els)

    def remove_alt(self, hashes, num_els=1):
        """ Remove an element by using the hash representation

            Args:
                hashes (list): The hashes representing the element to remove
                num_els (int): The number of times to remove the element
            Returns:
                int: The number of times the element was likely inserted \
                after the removal """
        res = list()
        for i, val in enumerate(hashes):
            t_bin = (val % self.width) + (i * self.width)
            self._bins[t_bin] -= num_els
            if self._bins[t_bin] < INT32_T_MIN:
                self._bins[t_bin] = INT32_T_MIN
            res.append(self._bins[t_bin])
        self.__elements_added -= num_els
        if self.elements_added < INT64_T_MIN:
            self.__elements_added = INT64_T_MIN

        return self.__query_method(sorted(res))

    def check(self, key):
        """Check number of times element 'key' is in the count-min sketch

        Args:
            key (str): The key to check the number of times inserted
        Returns:
            int: The number of times the element was likely inserted"""
        hashes = self.hashes(key)
        return self.check_alt(hashes)

    def check_alt(self, hashes):
        """ Check the count-min sketch for an element by using the hash \
            representation

            Args:
                hashes (list): The hashes representing the element to check
            Returns:
                int: The number of times the element was likely inserted """
        bins = self.__get_values_sorted(hashes)
        return self.__query_method(bins)

    def export(self, filepath):
        """ Export the count-min sketch to disk

            Args:
                filename (str): The filename to which the count-min sketch \
                will be written. """
        with open(filepath, "wb") as filepointer:
            # write out the bins
            rep = "i" * len(self._bins)
            filepointer.write(pack(rep, *self._bins))
            filepointer.write(pack("IIq", self.width, self.depth, self.elements_added))

    def join(self, second):
        """ Join two count-min sketchs into a single count-min sketch; the
            calling count-min sketch will have the resulting combined data

            Args:
                second (CountMinSketch): The count-min sketch to merge
            Raises:
                TypeError: When second is not either a :class:`CountMinSketch`,\
                    :class:`CountMeanSketch` or :class:`CountMeanMinSketch`
                CountMinSketchError: Raised when the count-min sketches are \
                    not of the same dimensions
            Note:
                The calling count-min sketch will contain the combined data
                once complete
        """
        if not isinstance(second, (CountMinSketch)):
            msg = "Unable to merge a count-min sketch with {}".format(type(second))
            raise TypeError(msg)
        if (
            (self.width != second.width)
            or (self.depth != second.depth)
            or (self.hashes("test") != second.hashes("test"))
        ):
            raise CountMinSketchError("Unable to merge as the count-min sketches are mismatched")

        size = self.width * self.depth
        for i in range(size):
            if self._bins[i] == INT32_T_MIN or self._bins[i] == INT32_T_MAX:
                continue
            self._bins[i] += second._bins[i]
            if self._bins[i] > INT32_T_MAX:
                self._bins[i] = INT32_T_MAX
            elif self._bins[i] < INT32_T_MIN:
                self._bins[i] = INT32_T_MIN

        # handle adding and removing elements added including handling overflow
        self.__elements_added += self.elements_added

        if self.elements_added > INT64_T_MAX:
            self.__elements_added = INT64_T_MAX
        elif self.elements_added < INT64_T_MIN:
            self.__elements_added = INT64_T_MIN

    def __load(self, filepath):
        """ load the count-min sketch from file """
        with open(filepath, "rb") as filepointer:
            offset = calcsize("IIq")
            filepointer.seek(offset * -1, os.SEEK_END)
            mybytes = unpack("IIq", filepointer.read(offset))
            self.__width = mybytes[0]
            self.__depth = mybytes[1]
            self.__elements_added = mybytes[2]
            self.__confidence = 1 - (1 / math.pow(2, self.depth))
            self.__error_rate = 2 / self.width

            filepointer.seek(0, os.SEEK_SET)
            length = self.width * self.depth
            rep = "i" * length
            offset = calcsize(rep)
            self._bins = list(unpack(rep, filepointer.read(offset)))

    def __get_values_sorted(self, hashes):
        """ get the values sorted """
        bins = list()
        for i, val in enumerate(hashes):
            t_bin = (val % self.width) + (i * self.width)
            bins.append(self._bins[t_bin])
        bins.sort()
        return bins

    @staticmethod
    def __min_query(results):
        """ generate the min query; assumes sorted list """
        return results[0]

    def __mean_query(self, results):
        """ generate the mean query; assumes sorted list """
        return sum(results) // self.depth

    def __mean_min_query(self, results):
        """ generate the mean-min query; assumes sorted list """
        if results[0] == 0 and results[-1] == 0:
            return 0
        meanmin = list()
        for t_bin in results:
            diff = self.elements_added - t_bin
            calc = t_bin - diff // (self.width - 1)
            meanmin.append(calc)
        meanmin.sort()
        if self.depth % 2 == 0:
            calc = meanmin[self.depth // 2] + meanmin[self.depth // 2 - 1]
            res = calc // 2
        else:
            res = meanmin[self.depth // 2]
        return res


class CountMeanSketch(CountMinSketch):
    """ Simple Count-Mean Sketch implementation for use in python;
        It can read and write the same format as the c version
        (https://github.com/barrust/count-min-sketch)

        Args:
            width (int): The width of the count-min sketch
            depth (int): The depth of the count-min sketch
            confidence (float): The level of confidence desired
            error_rate (float): The desired error rate
            filepath (str): Path to file to load
            hash_function (function): Hashing strategy function to use \
            `hf(key, number)`
        Returns:
            CountMeanSketch: A Count-Mean Sketch object
        Note:
            Initialization order of operations:
                1) From file
                2) Width and depth
                3) Confidence and error rate
        Note:
            Default query type is `mean`
        Note:
            For width and depth, width may realistically be in the thousands \
            while depth is in the single digit to teens  """

    __slots__ = CountMinSketch.__slots__

    def __init__(
        self,
        width=None,
        depth=None,
        confidence=None,
        error_rate=None,
        filepath=None,
        hash_function=None,
    ):
        super(CountMeanSketch, self).__init__(width, depth, confidence, error_rate, filepath, hash_function)
        self.query_type = "mean"


class CountMeanMinSketch(CountMinSketch):
    """ Simple Count-Mean-Min Sketch implementation for use in python;
        It can read and write the same format as the c version
        (https://github.com/barrust/count-min-sketch)

        Args:
            width (int): The width of the count-min sketch
            depth (int): The depth of the count-min sketch
            confidence (float): The level of confidence desired
            error_rate (float): The desired error rate
            filepath (str): Path to file to load
            hash_function (function): Hashing strategy function to use \
            `hf(key, number)`
        Returns:
            CountMeanMinSketch: A Count-Mean-Min Sketch object
        Note:
            Initialization order of operations:
                1) From file
                2) Width and depth
                3) Confidence and error rate
        Note:
            Default query type is `mean-min`
        Note:
            For width and depth, width may realistically be in the thousands \
            while depth is in the single digit to teens  """

    __slots__ = CountMinSketch.__slots__

    def __init__(
        self,
        width=None,
        depth=None,
        confidence=None,
        error_rate=None,
        filepath=None,
        hash_function=None,
    ):
        super(CountMeanMinSketch, self).__init__(width, depth, confidence, error_rate, filepath, hash_function)
        self.query_type = "mean-min"


class HeavyHitters(CountMinSketch):
    """ Find and track those elements that are the most common, or heavy
        hitters

        Args:
            num_hitters (int): The maximum number of distinct elements to track
            width (int): The width of the count-min sketch
            depth (int): The depth of the count-min sketch
            confidence (float): The level of confidence desired
            error_rate (float): The desired error rate
            filepath (str): Path to file to load
            hash_function (function): Hashing strategy function to use \
            `hf(key, number)`
        Returns:
            HeavyHitters: A Count-Min Sketch object
        Note:
            Initialization order of operations:
                1) From file
                2) Width and depth
                3) Confidence and error rate
        Note:
            Default query type is `min`
        Note:
            For width and depth, width may realistically be in the thousands \
            while depth is in the single digit to teens  """

    __slots__ = ["__top_x", "__top_x_size", "__num_hitters", "__smallest"]

    def __init__(
        self,
        num_hitters=100,
        width=None,
        depth=None,
        confidence=None,
        error_rate=None,
        filepath=None,
        hash_function=None,
    ):

        super(HeavyHitters, self).__init__(width, depth, confidence, error_rate, filepath, hash_function)
        self.__top_x = dict()  # top x heavy hitters
        self.__top_x_size = 0
        self.__num_hitters = num_hitters
        self.__smallest = 0

    def __str__(self):
        """ heavy hitters string rep """
        msg = super(HeavyHitters, self).__str__()
        tmp = "Heavy Hitters {0}\n" "\tNumber Hitters: {1}\n" "\tNumber Recorded: {2}"
        return tmp.format(msg, self.number_heavy_hitters, self.__top_x_size)

    @property
    def heavy_hitters(self):
        """dict: Return the heavy hitters, or most common elements

        Note:
            Not settable"""
        return self.__top_x

    @property
    def number_heavy_hitters(self):
        """int: Return the maximum number of heavy hitters being tracked

        Note:
            Not settable"""
        return self.__num_hitters

    def add(self, key, num_els=1):
        """Add element to heavy hitters

        Args:
            key (str): The element to add
            num_els (int): The number of instances to add
        Returns:
            int: Number of times key has been inserted
        Note:
            Override function"""
        hashes = self.hashes(key)
        return self.add_alt(key, hashes, num_els)

    def add_alt(self, key, hashes, num_els=1):
        """ Add the element `key` represented as hashes to the HeavyHitters
            object (hence the different signature on the function!)

            Args:
                key (str): The element to add
                hashes (list): The list of integers representing the key to \
                insert
                num_els (int): The number of instances to add
            Returns:
                int: Number of times key has been inserted
            Note:
                Different key signature than the normal :class:`CountMinSketch`
            Note:
                Override function """
        res = super(HeavyHitters, self).add_alt(hashes, num_els)

        # update the heavy hitters list as necessary
        if self.__top_x_size < self.__num_hitters:  # still have room in top x
            tmp = self.__top_x.get(key, None)
            self.__top_x[key] = res
            if tmp is None:
                self.__top_x_size = len(self.__top_x)
        elif key in self.__top_x:  # easy update as it is already there
            self.__top_x[key] = res
        elif res > self.__smallest:  # something in there is smaller
            self.__top_x[key] = res
            # get the key with the smallest element
            tmp_key = min(self.__top_x, key=self.__top_x.get)
            # delete this key
            self.__top_x.pop(tmp_key, None)
            new_min = min(self.__top_x, key=self.__top_x.get)
            self.__smallest = self.__top_x[new_min]
        return res

    def remove_alt(self, hashes, num_els=1):
        """ Remove element based on hashes provided; not supported in
            heavy hitters

            Raises:
                NotSupportedError: This function is not supported by the \
                HeavyHitters class
            Note:
                Override function """
        msg = (
            "Unable to remove elements in the HeavyHitters "
            "class as it is an un supported action (and does not"
            "make sense)!"
        )
        raise NotSupportedError(msg)

    def clear(self):
        """ Clear out the heavy hitters! """
        super(HeavyHitters, self).clear()
        self.__top_x = dict()
        self.__top_x_size = 0
        self.__smallest = 0

    def join(self, second):
        """ Join is not supported by HeavyHitters

            Raises:
                NotSupportedError: This functionality is currently not \
                supported """
        msg = "Joining is not supported for heavy hitters"
        raise NotSupportedError(msg)


class StreamThreshold(CountMinSketch):
    """ keep track of those elements over a certain threshold """

    __slots__ = ["__threshold", "__meets_threshold"]

    def __init__(
        self,
        threshold=100,
        width=None,
        depth=None,
        confidence=None,
        error_rate=None,
        filepath=None,
        hash_function=None,
    ):
        super(StreamThreshold, self).__init__(width, depth, confidence, error_rate, filepath, hash_function)
        self.__threshold = threshold
        self.__meets_threshold = dict()

    def __str__(self):
        """ stream threshold string rep """
        msg = super(StreamThreshold, self).__str__()
        tmp = "Stream Threshold {0}\n" "\tThreshold: {1}\n" "\tNumber Meeting Threshold: {2}"
        return tmp.format(msg, self.threshold, len(self.__meets_threshold))

    @property
    def meets_threshold(self):
        """ dict: Those keys that meet the required threshold (with value) """
        return self.__meets_threshold

    @property
    def threshold(self):
        """ int: The threshold at which a key is tracked """
        return self.__threshold

    def clear(self):
        """ Clear out the stream threshold! """
        super(StreamThreshold, self).clear()
        self.__meets_threshold = dict()

    def add(self, key, num_els=1):
        """Add the element for key into the data structure

        Args:
            key (str): The element to add
            num_els (int): The number of instances to add
        Returns:
            int: Number of times key has been inserted
        Note:
            Override function"""
        hashes = self.hashes(key)
        return self.add_alt(key, hashes, num_els)

    def add_alt(self, key, hashes, num_els=1):
        """ Add the element for key into the data structure

            Args:
                key (str): The element to add
                hashes (list): The list of integers representing the key to \
                insert
                num_els (int): The number of instances to add
            Returns:
                int: Number of times key has been inserted
            Note:
                Different key signature than the normal :class:`CountMinSketch`
            Note:
                Override function """
        res = super(StreamThreshold, self).add_alt(hashes, num_els)
        if res >= self.__threshold:
            self.__meets_threshold[key] = res
        return res

    def remove(self, key, num_els=1):
        """ Remove element 'key' from the count-min sketch

            Args:
                key (str): The element to remove
                num_els (int): The number of times to remove the element
            Returns:
                int: The number of times the element was likely inserted \
                after the removal
            Note:
                Override function """
        hashes = self.hashes(key)
        return self.remove_alt(key, hashes, num_els)

    def remove_alt(self, key, hashes, num_els=1):
        """ Remove an element by using the hash representation

            Args:
                key (str): The key that the hashes represent
                hashes (list): The hashes representing the element to remove
                num_els (int): The number of times to remove the element
            Returns:
                int: The number of times the element was likely inserted \
                after the removal
            Note:
                Different key signature than the normal :class:`CountMinSketch`
            Note:
                Override function """
        res = super(StreamThreshold, self).remove_alt(hashes, num_els)
        if res < self.__threshold:
            self.__meets_threshold.pop(key, None)
        else:
            self.__meets_threshold[key] = res
        return res

    def join(self, second):
        """ Join is not supported by StreamThreshold

            Raises:
                NotSupportedError: This functionality is currently not \
                supported """
        msg = "Joining is not supported for stream threshold"
        raise NotSupportedError(msg)



""" pyprobables module """

from .blooms import (
    BloomFilter,
    BloomFilterOnDisk,
    CountingBloomFilter,
    ExpandingBloomFilter,
    RotatingBloomFilter,
)
from .countminsketch import (
    CountMeanMinSketch,
    CountMeanSketch,
    CountMinSketch,
    HeavyHitters,
    StreamThreshold,
)
from .cuckoo import CountingCuckooFilter, CuckooFilter
from .exceptions import (
    CuckooFilterFullError,
    InitializationError,
    NotSupportedError,
    ProbablesBaseException,
    RotatingBloomFilterError,
)

__author__ = "Tyler Barrus"
__maintainer__ = "Tyler Barrus"
__email__ = "barrust@gmail.com"
__license__ = "MIT"
__version__ = "0.5.1"
__credits__ = []
__url__ = "https://github.com/barrust/pyprobables"
__bugtrack_url__ = "https://github.com/barrust/pyprobables/issues"

__all__ = [
    "BloomFilter",
    "BloomFilterOnDisk",
    "CountingBloomFilter",
    "CountMinSketch",
    "CountMeanSketch",
    "CountMeanMinSketch",
    "HeavyHitters",
    "StreamThreshold",
    "CuckooFilter",
    "CountingCuckooFilter",
    "InitializationError",
    "NotSupportedError",
    "ProbablesBaseException",
    "CuckooFilterFullError",
    "ExpandingBloomFilter",
    "RotatingBloomFilter",
    "RotatingBloomFilterError",
]



""" Probables Hashing library """

from functools import wraps
from hashlib import md5, sha256
from struct import unpack

from .constants import UINT64_T_MAX


def hash_with_depth_bytes(func):
    """Decorator to turns a function taking a single key and hashes it to
    bytes. Wraps functions to be used in Bloom filters and Count-Min sketch
    data structures.

    Args:
        key (str): The element to be hashed
        depth (int): The number of hash permutations to compute
    Returns:
        list(int): 64-bit hashed representation of key
    Note:
        Arguments shown are as it will be after decorated"""

    @wraps(func)
    def hashing_func(key, depth=1):
        """ wrapper function """
        res = list()
        tmp = key.encode("utf-8")
        for idx in range(depth):
            tmp = func(tmp, idx)
            res.append(unpack("Q", tmp[:8])[0])  # turn into 64 bit number
        return res

    return hashing_func


def hash_with_depth_int(func):
    """Decorator to turn a function that takes a single key and hashes it to
    an int. Wraps functions to be used in Bloom filters and Count-Min
    sketch data structures.

    Args:
        key (str): The element to be hashed
        depth (int): The number of hash permutations to compute
    Returns:
        list(int): 64-bit hashed representation of key
    Note:
        Arguments shown are as it will be after decorated"""

    @wraps(func)
    def hashing_func(key, depth=1):
        """ wrapper function """
        res = list()
        tmp = func(key, 0)
        res.append(tmp)
        for idx in range(1, depth):
            tmp = func("{0:x}".format(tmp), idx)
            res.append(tmp)
        return res

    return hashing_func


def default_fnv_1a(key, depth=1):
    """The default fnv-1a hashing routine

    Args:
        key (str): The element to be hashed
        depth (int): The number of hash permutations to compute
    Returns:
        list(int): List of size depth hashes"""

    res = list()
    for idx in range(depth):
        res.append(fnv_1a(key, idx))
    return res


def fnv_1a(key, seed=0):
    """Pure python implementation of the 64 bit fnv-1a hash

    Args:
        key (str): The element to be hashed
        seed (int): Add a seed to the initial starting point (0 means no seed)
    Returns:
        int: 64-bit hashed representation of key
    Note:
        Uses the lower 64 bits when overflows occur"""
    max64mod = UINT64_T_MAX + 1
    hval = (14695981039346656037 + (31 * seed)) % max64mod
    fnv_64_prime = 1099511628211
    tmp = map(ord, key)
    for t_str in tmp:
        hval ^= t_str
        hval *= fnv_64_prime
        hval %= max64mod
    return hval


@hash_with_depth_bytes
def default_md5(key, depth=1):
    """The default md5 hashing routine

    Args:
        key (str): The element to be hashed
        depth (int): The number of hash permutations to compute
    Returns:
        list(int): List of 64-bit hashed representation of key hashes
    Note:
        Returns the upper-most 64 bits"""
    return md5(key).digest()


@hash_with_depth_bytes
def default_sha256(key, depth=1):
    """The default sha256 hashing routine

    Args:
        key (str): The element to be hashed
        depth (int): The number of hash permutations to compute
    Returns:
        list(int): List of 64-bit hashed representation of key hashes
    Note:
        Returns the upper-most 64 bits"""
    return sha256(key).digest()



""" BloomFilter, python implementation
    License: MIT
    Author: Tyler Barrus (barrust@gmail.com)
    URL: https://github.com/barrust/bloom
"""

import mmap
import os
from shutil import copyfile
from struct import calcsize, pack, unpack

from ..exceptions import InitializationError, NotSupportedError
from ..utilities import is_hex_string, is_valid_file
from .basebloom import BaseBloom

MISMATCH_MSG = "The parameter second must be of type BloomFilter or " "a BloomFilterOnDisk"


def _verify_not_type_mismatch(second):
    """ verify that there is not a type mismatch """
    if not isinstance(second, (BloomFilter, BloomFilterOnDisk)):
        return False
    return True


def _cnt_set_bits(i):
    """ count number of bits set in this int """
    return bin(i).count("1")


def _tmp_jaccard_index(first, second):
    """ encapsulate the basics of the jaccard index """
    count_union = 0
    count_int = 0
    for i in list(range(0, first.bloom_length)):
        t_union = first._get_element(i) | second._get_element(i)
        t_intersection = first._get_element(i) & second._get_element(i)
        count_union += _cnt_set_bits(t_union)
        count_int += _cnt_set_bits(t_intersection)
    if count_union == 0:
        return 1.0
    return count_int / count_union


def _tmp_union(first, second):
    """ encapsulate the basics of the union """
    res = BloomFilter(
        first.estimated_elements,
        first.false_positive_rate,
        hash_function=first.hash_function,
    )
    for i in list(range(first.bloom_length)):
        res.bloom[i] = first._get_element(i) | second._get_element(i)
    res.elements_added = res.estimate_elements()
    return res


def _tmp_intersection(first, second):
    """ encapsulate the basics of the intersection """
    res = BloomFilter(
        first.estimated_elements,
        first.false_positive_rate,
        hash_function=first.hash_function,
    )

    for i in list(range(0, first.bloom_length)):
        res.bloom[i] = first._get_element(i) & second._get_element(i)
    res.elements_added = res.estimate_elements()
    return res


class BloomFilter(BaseBloom):
    """ Simple Bloom Filter implementation for use in python;
        It can read and write the same format as the c version
        (https://github.com/barrust/bloom)

        Args:
            est_elements (int): The number of estimated elements to be added
            false_positive_rate (float): The desired false positive rate
            filepath (str): Path to file to load
            hex_string (str): Hex based representation to be loaded
            hash_function (function): Hashing strategy function to use \
            `hf(key, number)`
        Returns:
            BloomFilter: A Bloom Filter object
        Note:
            Initialization order of operations:
                1) From file
                2) From Hex String
                3) From params """

    __slots__ = []

    def __init__(
        self,
        est_elements=None,
        false_positive_rate=None,
        filepath=None,
        hex_string=None,
        hash_function=None,
    ):
        """ setup the basic values needed """
        super(BloomFilter, self).__init__(
            "regular",
            est_elements,
            false_positive_rate,
            filepath,
            hex_string,
            hash_function,
        )

    def __str__(self):
        """ correctly handle python 3 vs python2 encoding if necessary """
        return self.__unicode__()

    def __unicode__(self):
        """ string / unicode representation of the Bloom Filter """
        on_disk = "no" if self.is_on_disk is False else "yes"
        stats = (
            "BloomFilter:\n"
            "\tbits: {0}\n"
            "\testimated elements: {1}\n"
            "\tnumber hashes: {2}\n"
            "\tmax false positive rate: {3:.6f}\n"
            "\tbloom length (8 bits): {4}\n"
            "\telements added: {5}\n"
            "\testimated elements added: {6}\n"
            "\tcurrent false positive rate: {7:.6f}\n"
            "\texport size (bytes): {8}\n"
            "\tnumber bits set: {9}\n"
            "\tis on disk: {10}\n"
        )
        return stats.format(
            self.number_bits,
            self.estimated_elements,
            self.number_hashes,
            self.false_positive_rate,
            self.bloom_length,
            self.elements_added,
            self.estimate_elements(),
            self.current_false_positive_rate(),
            self.export_size(),
            super(BloomFilter, self)._cnt_number_bits_set(),
            on_disk,
        )

    def intersection(self, second):
        """ Return a new Bloom Filter that contains the intersection of the
            two

            Args:
                second (BloomFilter): The Bloom Filter with which to take \
                the intersection
            Returns:
                BloomFilter: The new Bloom Filter containing the intersection
            Raises:
                TypeError: When second is not either a :class:`BloomFilter` \
                or :class:`BloomFilterOnDisk`
            Note:
                `second` may be a BloomFilterOnDisk object
            Note:
                If `second` is not of the same size (false_positive_rate and \
                est_elements) then this will return `None` """
        if not _verify_not_type_mismatch(second):
            raise TypeError(MISMATCH_MSG)

        if super(BloomFilter, self)._verify_bloom_similarity(second) is False:
            return None

        return _tmp_intersection(self, second)

    def union(self, second):
        """ Return a new Bloom Filter that contains the union of the two

            Args:
                second (BloomFilter): The Bloom Filter with which to \
                calculate the union
            Returns:
                BloomFilter: The new Bloom Filter containing the union
            Raises:
                TypeError: When second is not either a :class:`BloomFilter` \
                or :class:`BloomFilterOnDisk`
            Note:
                `second` may be a BloomFilterOnDisk object
            Note:
                If `second` is not of the same size (false_positive_rate and \
                est_elements) then this will return `None` """
        if not _verify_not_type_mismatch(second):
            raise TypeError(MISMATCH_MSG)

        if super(BloomFilter, self)._verify_bloom_similarity(second) is False:
            return None

        return _tmp_union(self, second)

    def jaccard_index(self, second):
        """ Calculate the jaccard similarity score between two Bloom Filters

            Args:
                second (BloomFilter): The Bloom Filter to compare with
            Returns:
                float: A numeric value between 0 and 1 where 1 is identical \
                and 0 means completely different
            Raises:
                TypeError: When second is not either a :class:`BloomFilter` \
                or :class:`BloomFilterOnDisk`
            Note:
                `second` may be a BloomFilterOnDisk object
            Note:
                If `second` is not of the same size (false_positive_rate and \
                est_elements) then this will return `None` """
        if not _verify_not_type_mismatch(second):
            raise TypeError(MISMATCH_MSG)

        if super(BloomFilter, self)._verify_bloom_similarity(second) is False:
            return None

        return _tmp_jaccard_index(self, second)

    def _cnt_number_bits_set(self):
        """ calculate the total number of set bits in the bloom """
        setbits = 0
        for i in list(range(0, self.bloom_length)):
            setbits += _cnt_set_bits(self._get_element(i))
        return setbits


class BloomFilterOnDisk(BaseBloom):
    """ Simple Bloom Filter implementation directly on disk for use in python;
        It can read and write the same format as the c version
        (https://github.com/barrust/bloom)

        Args:
            filepath (str): Path to file to load
            est_elements (int): The number of estimated elements to be added
            false_positive_rate (float): The desired false positive rate
            hex_string (str): Hex based representation to be loaded
            hash_function (function): Hashing strategy function to use \
            `hf(key, number)`
        Returns:
            BloomFilterOnDisk: A Bloom Filter object
        Raises:
            NotSupportedError: Loading using a hex string is not supported
        Note:
            Initialization order of operations:
                1) Esimated elements and false positive rate
                2) From Hex String
                3) Only filepath provided """

    __slots__ = ["__file_pointer", "__filename", "__export_offset"]

    def __init__(
        self,
        filepath,
        est_elements=None,
        false_positive_rate=None,
        hex_string=None,
        hash_function=None,
    ):
        # since we cannot load from a file only (to memory), we can't pass
        # the file to the constructor; therefore, we will have to catch
        # any exception thrown
        try:
            super(BloomFilterOnDisk, self).__init__(
                "reg-ondisk",
                est_elements=est_elements,
                false_positive_rate=false_positive_rate,
                hash_function=hash_function,
            )
        except InitializationError:
            pass

        self.__file_pointer = None
        self.__filename = None
        self.__export_offset = calcsize("Qf")
        self._on_disk = True

        if est_elements is not None and false_positive_rate is not None:
            # no need to check the file since this will over write it
            fpr = false_positive_rate
            vals = super(BloomFilterOnDisk, self)._set_optimized_params(est_elements, fpr, hash_function)
            super(BloomFilterOnDisk, self).__init__(
                "reg-ondisk",
                est_elements=est_elements,
                false_positive_rate=vals[1],
                hash_function=vals[0],
            )
            # do the on disk things
            with open(filepath, "wb") as filepointer:
                for _ in range(self.bloom_length):
                    filepointer.write(pack("B", int(0)))
                filepointer.write(pack("QQf", est_elements, 0, false_positive_rate))
                filepointer.flush()
            self.__load(filepath, hash_function)
        elif hex_string is not None and is_hex_string(hex_string):
            self._load_hex(hex_string, hash_function)
        elif is_valid_file(filepath):
            self.__load(filepath, hash_function)
        else:
            msg = "Insufecient parameters to set up the On Disk Bloom Filter"
            raise InitializationError(msg)

    def __del__(self):
        """ handle if user doesn't close the on disk Bloom Filter """
        self.close()

    def close(self):
        """ Clean up the BloomFilterOnDisk object """
        if self.__file_pointer is not None:
            self.__update()
            self._bloom.close()  # close the mmap
            self.__file_pointer.close()
            self.__file_pointer = None

    def __load(self, filepath, hash_function=None):
        """ load the Bloom Filter on disk """
        # read the file, set the optimal params
        # mmap everything
        with open(filepath, "r+b") as filepointer:
            offset = calcsize("QQf")
            filepointer.seek(offset * -1, os.SEEK_END)
            mybytes = unpack("QQf", filepointer.read(offset))
            vals = super(BloomFilterOnDisk, self)._set_optimized_params(mybytes[0], mybytes[2], hash_function)
        super(BloomFilterOnDisk, self).__init__(
            "reg-ondisk",
            est_elements=mybytes[0],
            false_positive_rate=vals[1],
            hash_function=vals[0],
        )
        self.__file_pointer = open(filepath, "r+b")
        self._bloom = mmap.mmap(self.__file_pointer.fileno(), 0)
        self._on_disk = True
        self.__filename = filepath

    def export(self, filename):
        """ Export to disk if a different location

            Args:
                filename (str): The filename to which the Bloom Filter will \
                be exported
            Note:
                Only exported if the filename is not the original filename """
        self.__update()
        if filename != self.__filename:
            # setup the new bloom filter
            copyfile(self.__filename, filename)
        # otherwise, nothing to do!

    def add_alt(self, hashes):
        super(BloomFilterOnDisk, self).add_alt(hashes)
        self.__update()

    def union(self, second):
        """ Return a new Bloom Filter that contains the union of the two

            Args:
                second (BloomFilter): The Bloom Filter with which to \
                calculate the union
            Returns:
                BloomFilter: The new Bloom Filter containing the union
            Raises:
                TypeError: When second is not either a :class:`BloomFilter` \
                or :class:`BloomFilterOnDisk`
            Note:
                `second` may be a BloomFilter object
            Note:
                If `second` is not of the same size (false_positive_rate and \
                est_elements) then this will return `None` """
        if not _verify_not_type_mismatch(second):
            raise TypeError(MISMATCH_MSG)

        if super(BloomFilterOnDisk, self)._verify_bloom_similarity(second) is False:
            return None

        return _tmp_union(self, second)

    def intersection(self, second):
        """ Return a new Bloom Filter that contains the intersection of the
            two

            Args:
                second (BloomFilter): The Bloom Filter with which to take \
                the intersection
            Returns:
                BloomFilter: The new Bloom Filter containing the intersection
            Raises:
                TypeError: When second is not either a :class:`BloomFilter` \
                or :class:`BloomFilterOnDisk`
            Note:
                `second` may be a BloomFilter object
            Note:
                If `second` is not of the same size (false_positive_rate and \
                est_elements) then this will return `None` """
        if not _verify_not_type_mismatch(second):
            raise TypeError(MISMATCH_MSG)

        if super(BloomFilterOnDisk, self)._verify_bloom_similarity(second) is False:
            return None

        return _tmp_intersection(self, second)

    def jaccard_index(self, second):
        """ Calculate the jaccard similarity score between two Bloom Filters

            Args:
                second (BloomFilter): The Bloom Filter to compare with
            Returns:
                float: A numeric value between 0 and 1 where 1 is identical \
                and 0 means completely different
            Raises:
                TypeError: When second is not either a :class:`BloomFilter` \
                or :class:`BloomFilterOnDisk`
            Note:
                `second` may be a BloomFilter object
            Note:
                If `second` is not of the same size (false_positive_rate and \
                est_elements) then this will return `None` """
        if not _verify_not_type_mismatch(second):
            raise TypeError(MISMATCH_MSG)

        if super(BloomFilterOnDisk, self)._verify_bloom_similarity(second) is False:
            return None
        return _tmp_jaccard_index(self, second)

    def export_hex(self):
        """ Export to a hex string

            Raises:
                NotSupportedError: This functionality is currently not \
                supported """
        msg = "`export_hex` is currently not supported by the on disk " "Bloom Filter"
        raise NotSupportedError(msg)

    def _load_hex(self, hex_string, hash_function=None):
        """ load from hex ... """
        msg = "Loading from hex_string is currently not supported by the " "on disk Bloom Filter"
        raise NotSupportedError(msg)

    def _get_element(self, idx):
        """ wrappper to use similar functions always! """
        return unpack("B", bytes([self._bloom[idx]]))[0]

    def __update(self):
        """update the on disk Bloom Filter and ensure everything is out
        to disk"""
        self._bloom.flush()
        self.__file_pointer.seek(-self.__export_offset, os.SEEK_END)
        self.__file_pointer.write(pack("Q", self.elements_added))
        self.__file_pointer.flush()

    def _cnt_number_bits_set(self):
        """ calculate the total number of set bits in the bloom """
        setbits = 0
        for i in list(range(0, self.bloom_length)):
            setbits += _cnt_set_bits(self._get_element(i))
        return setbits



""" PyProbables Exceptions """


class ProbablesBaseException(Exception):
    """Base ProbablesBaseException

    Args:
        message (str): The error message to be reported"""

    def __init__(self, message):
        self.message = message
        super(ProbablesBaseException, self).__init__(self.message)

    def __unicode__(self):
        return self.message

    def __str__(self):
        return self.__unicode__()


class InitializationError(ProbablesBaseException):
    """Initialization Exception

    Args:
        message (str): The initialization error messge"""

    def __init__(self, message):
        self.message = message
        super(InitializationError, self).__init__(self.message)


class NotSupportedError(ProbablesBaseException):
    """Not Supported Functionality Exception

    Args:
        message (str): The error message to be reported"""

    def __init__(self, message):
        self.message = message
        super(NotSupportedError, self).__init__(self.message)


class CuckooFilterFullError(ProbablesBaseException):
    """Cuckoo Filter Full Exception

    Args:
        message (str): The error message to be reported"""

    def __init__(self, message):
        self.message = message
        super(CuckooFilterFullError, self).__init__(self.message)


class RotatingBloomFilterError(ProbablesBaseException):
    """RotatingBloomFilter unable to rotate Blooms Exceptions

    Args:
        message (str): The error message to be reported"""

    def __init__(self, message):
        self.message = message
        super(RotatingBloomFilterError, self).__init__(self.message)


class CountMinSketchError(ProbablesBaseException):
    """CountMinSketch Exception

    Args:
        message (str): The error message to be reported"""

    def __init__(self, message):
        self.message = message
        super(CountMinSketchError, self).__init__(self.message)



""" Bloom Filter based data structures """

from .bloom import BloomFilter, BloomFilterOnDisk
from .countingbloom import CountingBloomFilter
from .expandingbloom import ExpandingBloomFilter, RotatingBloomFilter

__all__ = [
    "BloomFilter",
    "BloomFilterOnDisk",
    "CountingBloomFilter",
    "ExpandingBloomFilter",
    "RotatingBloomFilter",
]



""" CountingBloomFilter, python implementation
    License: MIT
    Author: Tyler Barrus (barrust@gmail.com)
    URL: https://github.com/barrust/counting_bloom
"""

from ..constants import UINT32_T_MAX, UINT64_T_MAX
from .basebloom import BaseBloom

MISMATCH_MSG = "The parameter second must be of type CountingBloomFilter"


def _verify_not_type_mismatch(second):
    """ verify that there is not a type mismatch """
    if not isinstance(second, (CountingBloomFilter)):
        return False
    return True


class CountingBloomFilter(BaseBloom):
    """ Simple Counting Bloom Filter implementation for use in python;
        It can read and write the same format as the c version
        (https://github.com/barrust/counting_bloom)

        Args:
            est_elements (int): The number of estimated elements to be added
            false_positive_rate (float): The desired false positive rate
            filepath (str): Path to file to load
            hex_string (str): Hex based representation to be loaded
            hash_function (function): Hashing strategy function to use \
            `hf(key, number)`
        Returns:
            CountingBloomFilter: A Counting Bloom Filter object

        Note:
            Initialization order of operations:
                1) From file
                2) From Hex String
                3) From params """

    __slots__ = BaseBloom.__slots__

    def __init__(
        self,
        est_elements=None,
        false_positive_rate=None,
        filepath=None,
        hex_string=None,
        hash_function=None,
    ):
        """ setup the basic values needed """
        super(CountingBloomFilter, self).__init__(
            "counting",
            est_elements=est_elements,
            false_positive_rate=false_positive_rate,
            filepath=filepath,
            hex_string=hex_string,
            hash_function=hash_function,
        )

    def __str__(self):
        """ correctly handle python 3 vs python2 encoding if necessary """
        return self.__unicode__()

    def __unicode__(self):
        """ string / unicode representation of the counting bloom filter """
        on_disk = "no" if self.is_on_disk is False else "yes"

        cnt = 0
        total = 0
        largest = 0
        largest_idx = 0
        for i, val in enumerate(self._bloom):
            total += val
            if val > 0:
                cnt += val
            if val > largest:
                largest = val
                largest_idx = i
        fullness = cnt / self.number_bits
        els_added = total // self.number_hashes

        stats = (
            "CountingBloom:\n"
            "\tbits: {0}\n"
            "\testimated elements: {1}\n"
            "\tnumber hashes: {2}\n"
            "\tmax false positive rate: {3:.6f}\n"
            "\telements added: {4}\n"
            "\tcurrent false positive rate: {5:.6f}\n"
            "\tis on disk: {6}\n"
            "\tindex fullness: {7:.6}\n"
            "\tmax index usage: {8}\n"
            "\tmax index id: {9}\n"
            "\tcalculated elements: {10}\n"
        )
        return stats.format(
            self.number_bits,
            self.estimated_elements,
            self.number_hashes,
            self.false_positive_rate,
            self.elements_added,
            self.current_false_positive_rate(),
            on_disk,
            fullness,
            largest,
            largest_idx,
            els_added,
        )

    def add(self, key, num_els=1):
        """Add the key to the Counting Bloom Filter

        Args:
            key (str): The element to be inserted
            num_els (int): Number of times to insert the element
        Returns:
            int: Maximum number of insertions"""
        hashes = self.hashes(key)
        return self.add_alt(hashes, num_els)

    def add_alt(self, hashes, num_els=1):
        """ Add the element represented by hashes into the Counting Bloom
            Filter

            Args:
                hashes (list): A list of integers representing the key to \
                insert
                num_els (int): Number of times to insert the element
            Returns:
                int: Maximum number of insertions """
        res = UINT32_T_MAX
        for i in list(range(0, self.number_hashes)):
            k = int(hashes[i]) % self.number_bits
            j = self._get_element(k)
            tmp = j + num_els
            if tmp <= UINT32_T_MAX:
                self.bloom[k] = j + num_els
            else:
                self.bloom[k] = UINT32_T_MAX
            if self.bloom[k] < res:
                res = self.bloom[k]
        self.elements_added += num_els
        if self.elements_added > UINT64_T_MAX:
            self.elements_added = UINT64_T_MAX
        return res

    def check(self, key):
        """Check if the key is likely in the Counting Bloom Filter

        Args:
            key (str): The element to be checked
        Returns:
            int: Maximum number of insertions"""
        hashes = self.hashes(key)
        return self.check_alt(hashes)

    def check_alt(self, hashes):
        """ Check if the element represented by hashes is in the Counting
            Bloom Filter

            Args:
                hashes (list): A list of integers representing the key to \
                check
            Returns:
                int: Maximum number of insertions """
        res = UINT32_T_MAX
        for i in list(range(0, self.number_hashes)):
            k = int(hashes[i]) % self.number_bits
            tmp = self._get_element(k)
            if tmp < res:
                res = tmp
        return res

    def remove(self, key, num_els=1):
        """Remove the element from the counting bloom

        Args:
            key (str): The element to be removed
            num_els (int): Number of times to remove the element
        Returns:
            int: Maximum number of insertions after the removal"""
        hashes = self.hashes(key)
        return self.remove_alt(hashes, num_els)

    def remove_alt(self, hashes, num_els=1):
        """ Remvoe the element represented by hashes from the Counting Bloom \
            Filter

            Args:
                hashes (list): A list of integers representing the key to \
                remove
                num_els (int): Number of times to remove the element
            Returns:
                int: Maximum number of insertions after the removal """
        tmp = self.check_alt(hashes)
        if tmp == UINT32_T_MAX:  # cannot remove if we have hit the max
            return UINT32_T_MAX
        elif tmp == 0:
            return 0

        # determine how many we can actually remove
        if tmp - num_els < 0:
            t_num_els = tmp
        else:
            t_num_els = num_els
        for i in list(range(self.number_hashes)):
            k = int(hashes[i]) % self.number_bits
            j = self._get_element(k)
            self.bloom[k] = j - t_num_els
        self.elements_added -= t_num_els
        return tmp - t_num_els

    def intersection(self, second):
        """ Take the intersection of two Counting Bloom Filters

            Args:
                second (CountingBloomFilter): The Bloom Filter with which to \
                take the intersection
            Returns:
                CountingBloomFilter: The new Counting Bloom Filter containing \
                the union
            Raises:
                TypeError: When second is not a :class:`CountingBloomFilter`
            Note:
                The elements_added property will be set to the estimated \
                number of unique elements added as found in \
                estimate_elements()
            Note:
                If `second` is not of the same size (false_positive_rate and \
                est_elements) then this will return `None` """
        if not _verify_not_type_mismatch(second):
            raise TypeError(MISMATCH_MSG)

        if super(CountingBloomFilter, self)._verify_bloom_similarity(second) is False:
            return None
        res = CountingBloomFilter(
            est_elements=self.estimated_elements,
            false_positive_rate=self.false_positive_rate,
            hash_function=self.hash_function,
        )

        for i in list(range(self.bloom_length)):
            if self._get_element(i) > 0 and second._get_element(i) > 0:
                tmp = self._get_element(i) + second._get_element(i)
                res.bloom[i] = tmp
        res.elements_added = res.estimate_elements()
        return res

    def jaccard_index(self, second):
        """ Take the Jaccard Index of two Counting Bloom Filters

            Args:
                second (CountingBloomFilter): The Bloom Filter with which to \
                take the jaccard index
            Returns:
                float: A numeric value between 0 and 1 where 1 is identical \
                and 0 means completely different
            Raises:
                TypeError: When second is not a :class:`CountingBloomFilter`
            Note:
                The Jaccard Index is based on the unique set of elements \
                added and not the number of each element added
            Note:
                If `second` is not of the same size (false_positive_rate and \
                est_elements) then this will return `None` """
        if not _verify_not_type_mismatch(second):
            raise TypeError(MISMATCH_MSG)

        if super(CountingBloomFilter, self)._verify_bloom_similarity(second) is False:
            return None

        count_union = 0
        count_inter = 0
        for i in list(range(self.bloom_length)):
            if self._get_element(i) > 0 or second._get_element(i) > 0:
                count_union += 1
            if self._get_element(i) > 0 and second._get_element(i) > 0:
                count_inter += 1
        if count_union == 0:
            return 1.0
        return count_inter / count_union

    def union(self, second):
        """ Return a new Countiong Bloom Filter that contains the union of
            the two

            Args:
                second (CountingBloomFilter): The Counting Bloom Filter with \
                which to calculate the union
            Returns:
                CountingBloomFilter: The new Counting Bloom Filter containing \
                the union
            Raises:
                TypeError: When second is not a :class:`CountingBloomFilter`
            Note:
                The elements_added property will be set to the estimated \
                number of unique elements added as found in \
                estimate_elements()
            Note:
                If `second` is not of the same size (false_positive_rate and \
                est_elements) then this will return `None` """
        if not _verify_not_type_mismatch(second):
            raise TypeError(MISMATCH_MSG)

        if super(CountingBloomFilter, self)._verify_bloom_similarity(second) is False:
            return None
        res = CountingBloomFilter(
            est_elements=self.estimated_elements,
            false_positive_rate=self.false_positive_rate,
            hash_function=self.hash_function,
        )
        for i in list(range(self.bloom_length)):
            tmp = self._get_element(i) + second._get_element(i)
            res.bloom[i] = tmp
        res.elements_added = res.estimate_elements()
        return res

    def _cnt_number_bits_set(self):
        """ calculate the total number of set bits in the bloom """
        cnt = 0
        for i in list(range(self.bloom_length)):
            if self._get_element(i) > 0:
                cnt += 1
        return cnt



""" Module Installation script """
import io

import setuptools

from probables import (
    __author__,
    __bugtrack_url__,
    __email__,
    __license__,
    __url__,
    __version__,
)


def read_file(filepath):
    """ read the file """
    with io.open(filepath, "r") as filepointer:
        res = filepointer.read()
    return res


KEYWORDS = [
    "python",
    "probabilistic",
    "data-structure",
    "bloom",
    "filter",
    "count-min",
    "sketch",
    "bloom-filter",
    "count-min-sketch",
    "cuckoo-filter",
]

setuptools.setup(
    name="pyprobables",
    version=__version__,
    author=__author__,
    author_email=__email__,
    description="Probabilistic data structures in python",
    license=__license__,
    keywords=" ".join(KEYWORDS),
    url=__url__,
    download_url="{0}/tarball/v{1}".format(__url__, __version__),
    bugtrack_url=__bugtrack_url__,
    install_requires=read_file("requirements.txt").splitlines(),
    packages=setuptools.find_packages(include=["probables", "probables.*"]),
    long_description=read_file("README.rst"),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Topic :: Software Development :: Libraries",
        "Topic :: Utilities",
        "License :: OSI Approved",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    test_suite="tests",
)



""" count-min sketch submodule """

from .countminsketch import (
    CountMeanMinSketch,
    CountMeanSketch,
    CountMinSketch,
    HeavyHitters,
    StreamThreshold,
)

__all__ = [
    "CountMinSketch",
    "HeavyHitters",
    "StreamThreshold",
    "CountMeanSketch",
    "CountMeanMinSketch",
]



""" Counting Cuckoo Filter, python implementation
    License: MIT
    Author: Tyler Barrus (barrust@gmail.com)
"""

import os
import random
from struct import calcsize, pack, unpack

from ..exceptions import CuckooFilterFullError
from .cuckoo import CuckooFilter


class CountingCuckooFilter(CuckooFilter):
    """ Simple Counting Cuckoo Filter implementation

        Args:
            capacity (int): The number of bins
            bucket_size (int): The number of buckets per bin
            max_swaps (int): The number of cuckoo swaps before stopping
            expansion_rate (int): The rate at which to expand
            auto_expand (bool): If the filter should automatically expand
            finger_size (int): The size of the fingerprint to use in bytes \
            (between 1 and 4); exported as 4 bytes; up to the user to reset \
            the size correctly on import
            filename (str): The path to the file to load or None if no file
        Returns:
            CountingCuckooFilter: A Cuckoo Filter object """

    __slots__ = [
        "__unique_elements",
        "_inserted_elements",
        "_bucket_size",
        "__max_cuckoo_swaps",
        "_cuckoo_capacity",
        "_buckets",
    ]

    def __init__(
        self,
        capacity=10000,
        bucket_size=4,
        max_swaps=500,
        expansion_rate=2,
        auto_expand=True,
        finger_size=4,
        filepath=None,
        hash_function=None,
    ):
        """ setup the data structure """
        self.__unique_elements = 0
        super(CountingCuckooFilter, self).__init__(
            capacity,
            bucket_size,
            max_swaps,
            expansion_rate,
            auto_expand,
            finger_size,
            filepath,
            hash_function,
        )

    def __contains__(self, val):
        """ setup the `in` keyword """
        if self.check(val) > 0:
            return True
        return False

    @property
    def unique_elements(self):
        """ int: unique number of elements inserted """
        return self.__unique_elements

    def load_factor(self):
        """ float: How full the Cuckoo Filter is currently """
        return self.unique_elements / (self.capacity * self.bucket_size)

    def add(self, key):
        """ Add element key to the filter

            Args:
                key (str): The element to add
            Raises:
                CuckooFilterFullError: When element not inserted after \
                maximum number of swaps or 'kicks' """
        idx_1, idx_2, fingerprint = self._generate_fingerprint_info(key)

        is_present = self._check_if_present(idx_1, idx_2, fingerprint)
        if is_present is not None:
            for bucket in self.buckets[is_present]:
                if fingerprint in bucket:
                    bucket.increment()
                    self._inserted_elements += 1
                    return
        finger = self._insert_fingerprint_alt(fingerprint, idx_1, idx_2)
        self._deal_with_insertion(finger)

    def check(self, key):
        """Check if an element is in the filter

        Args:
            key (str): Element to check
        Returns:
            int: The number of times inserted into the filter"""
        idx_1, idx_2, fingerprint = self._generate_fingerprint_info(key)
        is_present = self._check_if_present(idx_1, idx_2, fingerprint)
        if is_present is not None:
            # get the count out!
            for bucket in self.buckets[is_present]:
                if fingerprint in bucket:
                    return bucket.count
        return 0

    def remove(self, key):
        """Remove an element from the filter

        Args:
            key (str): Element to remove"""
        idx_1, idx_2, fingerprint = self._generate_fingerprint_info(key)
        idx = self._check_if_present(idx_1, idx_2, fingerprint)
        if idx is None:
            return False
        for bucket in self.buckets[idx]:
            if fingerprint in bucket:
                bucket.decrement()
                self._inserted_elements -= 1
                if bucket.count == 0:
                    self.buckets[idx].remove(bucket)
                    self.__unique_elements -= 1
                return True
        return False  # catch this...

    def expand(self):
        """ Expand the cuckoo filter """
        self._expand_logic(None)

    def export(self, filename):
        """Export cuckoo filter to file

        Args:
            filename (str): Path to file to export"""
        with open(filename, "wb") as filepointer:
            for bucket in self.buckets:
                # do something for each...
                rep = len(bucket) * "II"
                wbyt = pack(rep, *[x for x in self.__bucket_decomposition(bucket)])
                filepointer.write(wbyt)
                leftover = self.bucket_size - len(bucket)
                rep = leftover * "II"
                filepointer.write(pack(rep, *([0] * (leftover * 2))))
            # now put out the required information at the end
            filepointer.write(pack("II", self.bucket_size, self.max_swaps))

    def _insert_fingerprint_alt(self, fingerprint, idx_1, idx_2, count=1):
        """ insert a fingerprint, but with a count parameter! """
        if self.__insert_element(fingerprint, idx_1, count):
            self._inserted_elements += 1
            self.__unique_elements += 1
            return None
        elif self.__insert_element(fingerprint, idx_2, count):
            self._inserted_elements += 1
            self.__unique_elements += 1
            return None

        # we didn't insert, so now we need to randomly select one index to use
        # and move things around to the other index, if possible, until we
        # either move everything around or hit the maximum number of swaps
        idx = random.choice([idx_1, idx_2])
        prv_bin = CountingCuckooBin(fingerprint, 1)
        for _ in range(self.max_swaps):
            # select one element to be swapped out...
            swap_elm = random.randint(0, self.bucket_size - 1)
            swap_finger = self.buckets[idx][swap_elm]
            prv_bin, self.buckets[idx][swap_elm] = swap_finger, prv_bin

            # now find another place to put this fingerprint
            index_1, index_2 = self._indicies_from_fingerprint(prv_bin.finger)

            idx = index_2 if idx == index_1 else index_1

            if self.__insert_element(prv_bin.finger, idx, prv_bin.count):
                self._inserted_elements += 1
                self.__unique_elements += 1
                return None

        # if we got here we have an error... we might need to know what is left
        return prv_bin

    def _check_if_present(self, idx_1, idx_2, fingerprint):
        """ wrapper for checking if fingerprint is already inserted """
        if fingerprint in [x.finger for x in self.buckets[idx_1]]:
            return idx_1
        elif fingerprint in [x.finger for x in self.buckets[idx_2]]:
            return idx_2
        return None

    def _load(self, filename):
        """ load a cuckoo filter from file """
        with open(filename, "rb") as filepointer:
            offset = calcsize("II")
            int_size = calcsize("II")
            filepointer.seek(offset * -1, os.SEEK_END)
            list_size = filepointer.tell()
            mybytes = unpack("II", filepointer.read(offset))
            self._bucket_size = mybytes[0]
            self.__max_cuckoo_swaps = mybytes[1]
            self._cuckoo_capacity = list_size // int_size // self.bucket_size
            self._inserted_elements = 0
            # now pull everything in!
            filepointer.seek(0, os.SEEK_SET)
            self._buckets = list()
            for i in range(self.capacity):
                self.buckets.append(list())
                for _ in range(self.bucket_size):
                    finger, count = unpack("II", filepointer.read(int_size))
                    if finger > 0:
                        ccb = CountingCuckooBin(finger, count)
                        self.buckets[i].append(ccb)
                        self._inserted_elements += count
                        self.__unique_elements += 1

    def _expand_logic(self, extra_fingerprint):
        """ the logic to acutally expand the cuckoo filter """
        # get all the fingerprints
        fingerprints = self._setup_expand(extra_fingerprint)
        self.__unique_elements = 0  # this needs to be reset!

        for elm in fingerprints:
            idx_1, idx_2 = self._indicies_from_fingerprint(elm.finger)
            res = self._insert_fingerprint_alt(elm.finger, idx_1, idx_2, elm.count)
            if res is not None:  # again, this *shouldn't* happen
                msg = "The CountingCuckooFilter failed to expand"
                raise CuckooFilterFullError(msg)

    def __insert_element(self, fingerprint, idx, count=1):
        """ insert an element """
        if len(self.buckets[idx]) < self.bucket_size:
            self.buckets[idx].append(CountingCuckooBin(fingerprint, count))
            return True
        return False

    @staticmethod
    def __bucket_decomposition(bucket):
        for buck in bucket:
            yield buck.finger
            yield buck.count


class CountingCuckooBin(object):
    """ A container class for the counting cuckoo filter """

    # keep it lightweight
    __slots__ = ["__fingerprint", "__count"]

    def __init__(self, fingerprint, count):
        """ init """
        self.__fingerprint = fingerprint
        self.__count = count

    def __contains__(self, val):
        """ setup the `in` construct """
        return self.__fingerprint == val

    @property
    def finger(self):
        """ fingerprint property """
        return self.__fingerprint

    @property
    def count(self):
        """ count property """
        return self.__count

    def __repr__(self):
        """ how do we represent this? """
        return self.__str__()

    def __str__(self):
        """ convert it into a string """
        return "(fingerprint:{} count:{})".format(self.__fingerprint, self.__count)

    def increment(self):
        """ increment """
        self.__count += 1
        return self.__count

    def decrement(self):
        """ decrement """
        self.__count -= 1
        return self.__count
